/*	$Id: ueagle_mod.c,v 1.1 2004/08/29 19:07:56 damien Exp $	*/

/*-
 * Copyright (c) 2005 Matthieu Castet <castet.matthieu@free.fr>
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#define EAGLEUSBVERSION "ueagle"

#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/init.h>
#include <linux/proc_fs.h>
#include <linux/crc32.h>
#include <linux/usb.h>
#include <linux/firmware.h>
#include <linux/ctype.h>

#include "eagle-usb.h"
#include "debug.h"

#include "ueaglevar.h"
#include "ueaglereg.h"


/*
 * Hardcoded endpoint addresses
 */
#define EP_BULK_DATA_OUT       0x02
#define EP_ISOC_DATA_IN        0x88
#define EP_BULK_IDMA_OUT       0x04
#define EP_INT_IN              0x84

#define USB_INTF_IN            0x02
#define FASTEST_ISO_INTF       0x08
#define FASTEST_ISO_RATE       1007

#define USBDEVNAME(x)  "ueagle"
#define DPRINTF(x,...) printk x
#define DPRINTFN(x,...) if(x < 4) printk __VA_ARGS__
#define printf printk
#define UPDATE_ATM_STAT(type, val) if (sc->usbatm->atm_dev) \
					sc->usbatm->atm_dev->type = val;

/*
#ifdef UEAGLE_DEBUG
#define DPRINTF(x)	if (ueagledebug > 0) logprintf x
#define DPRINTFN(n, x)	if (ueagledebug > (n)) logprintf x
int ueagledebug = 0;
#else
#define DPRINTF(x)
#define DPRINTFN(n, x)
#endif
*/
#define MSEC_TO_JIFFIES(_msec_) ((HZ*(unsigned long)_msec_)/1000)
#define wakeup(x) wake_up_interruptible(&x->sync_q);

#define usbd_status int

static int eu_read_proc(char *page, char **start, off_t off, int count,
			int *eof, void *data);

static struct usb_driver eu_driver;
/*
 * Our /proc dir entry
 */
static struct proc_dir_entry *eu_procdir;


/*
 * User supplied debug level
 */
unsigned int module_dbg_mask = 0x0;


static int dev = 0;
#define NB_MODEM 4
static char *dsp[NB_MODEM] = {[0 ... (NB_MODEM - 1)] = "DSP.bin"};
static char *cmv[NB_MODEM] = {[0 ... (NB_MODEM - 1)] = "CMV.bin"};
static int iso[NB_MODEM] = {[0 ... (NB_MODEM - 1)] = 1};
static int wait[NB_MODEM] = {[0 ... (NB_MODEM - 1)] = 0};


module_param(module_dbg_mask, uint, 0644);
MODULE_PARM_DESC(module_dbg_mask, "Module Debug mask");
module_param_array(dsp, charp, NULL, 0644);
MODULE_PARM_DESC(dsp, "dsp file");
module_param_array(cmv, charp, NULL, 0644);
MODULE_PARM_DESC(cmv, "cmv file");
module_param_array(iso, bool, NULL, 0644);
MODULE_PARM_DESC(iso, "iso mode");
module_param_array(wait, bool, NULL, 0644);
MODULE_PARM_DESC(wait, "wait the synchronisation before starting ATM");


/**
 * htoi - Converts hex to integer.
 *
 */
static uint8_t htoi ( uint8_t h ) 
{
    uint8_t d;
    
    if ( isxdigit (h) ) 
    {
        /*
         * This is an hexa-digit
         */
        d = isdigit (h) ? h - '0' : toupper (h) -'A'+ 10 ;
    }
    else
    {
        d = 0;
    }

    return (d);
}

static int ueagle_getesi(struct ueagle_softc *sc, u_char *esi)
{
	int maclen, i;
	uint8_t macbuf[ETH_ALEN*2+1];

	maclen = usb_string(sc->sc_dev, UEAGLE_ESISTRING, macbuf, ETH_ALEN*2+1);
	/*
	 * Make sure we got a string of the right length
	 */
	if ( maclen != ETH_ALEN*2 )
		return 1;
 
	/*
	 * Convert the string into a numeric format
	 * (each string byte is a nibble in the result)
	 */
	for ( i=0; i<ETH_ALEN; i++ )
	{
		esi[i] = (htoi(macbuf[i*2])<<4) + (htoi(macbuf[i*2+1]));
	}
	return 0;
}
#include "ueagle-bsd.c"

/* Firmware loading */
#define LOAD_INTERNAL     0xA0
#define F8051_USBCS       0x7f92
/**
 * eu_send_modem_cmd - Send a command to the modem (firmware loading)
 *
 */
static int eu_send_modem_cmd(struct usb_device *usb,
			     uint16_t addr, uint16_t count, uint8_t * buff)
{
	int ret = -ENOMEM;
	uint8_t *xfer_buff = NULL;

	xfer_buff = kmalloc(count, GFP_KERNEL);
	if (xfer_buff) {
		memcpy(xfer_buff, buff, count);
		ret = usb_control_msg(usb,
				      usb_sndctrlpipe(usb, 0),
				      LOAD_INTERNAL,
				      USB_DIR_OUT | USB_TYPE_VENDOR |
				      USB_RECIP_DEVICE, addr, 0, xfer_buff,
				      count, 300);
		kfree(xfer_buff);
	}
	return ret;
}

/**
 * eu_load_firmware - Load firmware into pre-firmware devices.
 *
 */
static int eu_load_firmware(struct usb_device *usb, unsigned int ver)
{
	int ret = -ENODEV;
	const struct firmware *fw_entry;
	int size;
	uint8_t *pfw;
	char *fw_name = "eagle.fw";
	uint32_t crc = 0;
	uint8_t value;

	eu_enters(DBG_UTILS);

	/*
	 * Send reset
	 */
	value = 1;
	ret = eu_send_modem_cmd(usb, F8051_USBCS, sizeof(value), &value);

	if (ret < 0) {
		eu_err("eu_send_modem_cmd reset failure with error %d\n", ret);
		goto byebye;
	}
	switch (ver) {
	case 0:
		fw_name = "elsa.fw";
		break;
	case 1:
		fw_name = "eagleI.fw";
		break;
	case 2:
		fw_name = "eagleII.fw";
		break;
	case 3:
		fw_name = "eagleIII.fw";
		break;
	}

	ret = request_firmware(&fw_entry, fw_name, &usb->dev);
	if (ret) {
		eu_err("%s : Firmware not available\n", fw_name);
		goto byebye;
	}
	pfw = fw_entry->data;
	size = fw_entry->size;

	crc = le32_to_cpu(*(uint32_t *) pfw);
	pfw += 4;
	size -= 4;
	if (crc32_be(0, pfw, size) != crc) {
		eu_err("%s : Firmware is corrupted\n", fw_name);
		release_firmware(fw_entry);
		ret = -1;
		goto byebye;
	}

	while (size > 0) {
		uint8_t len = *pfw;
		uint16_t add = le16_to_cpu(*(uint16_t *) (pfw + 1));
		ret = eu_send_modem_cmd(usb, add, len, pfw + 3);
		if (ret < 0) {
			eu_err
			    ("eu_send_modem_cmd of download data failed (%d)\n",
			     ret);
			goto byebye;
		}
		pfw += len + 3;
		size -= len + 3;
	}
	release_firmware(fw_entry);

	/*
	 * De-assert reset
	 */
	value = 0;
	ret = eu_send_modem_cmd(usb, F8051_USBCS, 1, &value);
	if (ret < 0) {
		eu_err("eu_send_modem_cmd de-reset failure (%d)\n", ret);
		goto byebye;
	}
	ret = 0;

      byebye:
	eu_leaves(DBG_UTILS);
	return (ret);
}

/**
 * eu_load_firmware - Load firmware into pre-firmware devices.
 *
 */
static int loadXILINXFirmware(struct ueagle_softc *sc)
{
	int ret = -ENODEV;
	const struct firmware *fw_entry;
	int size, u;
	uint8_t *pfw;
	char *fw_name = "elsa-pst.fw";
	uint32_t crc = 0;

	eu_enters(DBG_UTILS);

	ret = request_firmware(&fw_entry, fw_name, &sc->sc_dev->dev);
	if (ret) {
		eu_err("%s : Firmware not available\n", fw_name);
		goto byebye;
	}
	pfw = fw_entry->data;
	size = fw_entry->size;

	crc = le32_to_cpu(*(uint32_t *) pfw);
	pfw += 4;
	size -= 4;
	if (crc32_be(0, pfw, size) != crc) {
		eu_err("%s : Firmware is corrupted\n", fw_name);
		release_firmware(fw_entry);
		ret = -1;
		goto byebye;
	}

	for (u = 0; u < size; ) {
		int ln = size - u;
		if (ln > 64)
			ln = 64;
		ret = ueagle_request(sc, 0xe, 0, ln, pfw);
		if (ret < 0) {
			eu_err
			    ("elda download data failed (%d)\n",
			     ret);
			goto byebye;
		}
		pfw += u;
	}
	release_firmware(fw_entry);

	ret = ueagle_request(sc, 0xe, 1, 0, NULL);
	if (ret < 0) {
		eu_err("elsa de-reset failure (%d)\n", ret);
		goto byebye;
	}
	ret = 0;

      byebye:
	eu_leaves(DBG_UTILS);
	return (ret);
}
/* syfs interface */
static ssize_t read_status(struct device *dev, char *buf)
{
	struct usb_interface *intf = to_usb_interface(dev);
	struct usbatm_data * usbatm = usb_get_intfdata(intf);
	struct ueagle_softc *sc = (struct ueagle_softc *)usbatm->driver_data;
	/*int p = 0;
	switch ((sc->stats.phy.state >> 8) & 0xf) {
	case 1:
		p = sprintf(buf, "initializing\n");
		break;
	case 0:
		p = sprintf(buf, "booting\n");
		break;
	case 2:
		p = sprintf(buf, "operational\n");
		break;
	default:
		p = sprintf(buf, "failed\n");
		break;
	}
	return p;*/
	return sprintf(buf, "%8.8x\n", sc->stats.phy.state);
}

static ssize_t reboot(struct device *dev, const char *buf, size_t count)
{
	struct usb_interface *intf = to_usb_interface(dev);
	struct usbatm_data *usbatm = usb_get_intfdata(intf);
	struct ueagle_softc *sc = (struct ueagle_softc *)usbatm->driver_data;
	ueagle_startreset(sc);
	return count;
}

static DEVICE_ATTR(status, S_IWUGO | S_IRUGO, read_status, reboot);

#define UEAGLEATR(name,reset) static ssize_t read_##name(struct device *dev, char *buf)\
{\
	struct usb_interface *intf = to_usb_interface(dev);\
	struct usbatm_data * usbatm = usb_get_intfdata(intf);\
	struct ueagle_softc *sc = (struct ueagle_softc *)usbatm->driver_data;\
	int p; \
	p = sprintf(buf, "%8.8x\n", sc->stats.phy.name);\
	if (reset) \
	  sc->stats.phy.name = 0; \
	return p; \
}\
\
static DEVICE_ATTR(name, S_IRUGO, read_##name, NULL)

UEAGLEATR(mflags,1);


/* ATM stuff */
static int eu_atm_open(struct usbatm_data *usbatm, struct atm_dev *atm_dev)
{
	struct ueagle_softc *sc = (struct ueagle_softc *)usbatm->driver_data;

	//if (((sc->stats.phy.state >> 8) & 0xf) == 2)
		//ueagle_request(sc, UEAGLE_SETMODE, UEAGLE_LOOPBACKOFF, 0, NULL);
	sc->running = 1;

	atm_dev->signal = ATM_PHY_SIG_LOST;
	return ueagle_getesi(sc, atm_dev->esi);
}

static void eu_atm_stop(struct usbatm_data *usbatm, struct atm_dev *atm_dev)
{
	struct ueagle_softc *sc = (struct ueagle_softc *)usbatm->driver_data;

	//ueagle_request(sc, UEAGLE_SETMODE, UEAGLE_LOOPBACKON, 0, NULL);
	sc->running = 0;
}

static int eu_heavy(struct usbatm_data *usbatm, struct usb_interface *intf)
{
	struct ueagle_softc *sc = (struct ueagle_softc *)usbatm->driver_data;
	if (wait_event_interruptible(sc->sync_q ,
				((sc->stats.phy.state >> 8) & 0xf) == 2 ))
		return -EINTR;

	return 0;

}

static int eu_bind(struct usbatm_data *usbatm, struct usb_interface *intf,
		   const struct usb_device_id *id, int *heavy)
{
	char path[32];
	struct proc_dir_entry *procfile;
	int ret = 0;

	struct usb_interface *intf1;
	struct usb_device *usb = interface_to_usbdev(intf);
	struct ueagle_softc *sc;
	int ifnum = intf->altsetting->desc.bInterfaceNumber;

	eu_enters(DBG_INIT);

	if (ifnum != 0)
		return -ENODEV;

	*heavy = wait[dev];

	/*
	 * Interface 1 is for outbound traffic
	 */
	if (!(intf1 = usb_ifnum_to_if(usb, 1))) {
		dbg("interface 1 not found!");
		return -ENODEV;
	}

	ret = usb_driver_claim_interface(&eu_driver, intf1, usbatm);
	if (ret != 0) {
		eu_report("Failed to claim interface 1 (%d)\n", ret);
		return -EBUSY;
	}

	/*
	 * Interface 2 is for inbound traffic
	 */
	if (!(intf1 = usb_ifnum_to_if(usb, 2))) {
		dbg("interface 2 not found!");
		return -ENODEV;
	}

	ret = usb_driver_claim_interface(&eu_driver, intf1, usbatm);
	if (ret != 0) {
		eu_report("Failed to claim interface 2 (%d)\n", ret);
		return -EBUSY;
	}

	sc = kmalloc(sizeof(struct ueagle_softc), GFP_KERNEL);

	if (!sc) {
		eu_err("eu_init: Not enough memory !\n");
		return -ENOMEM;

	}

	sc->sc_dev = usb;
	usbatm->driver_data = sc;
	sc->usbatm = usbatm;
	sc->dev = dev++;
	sc->driver_info = id->driver_info & 0xf;

	init(sc);

	if (!(sc->driver_info & 0xf))
		loadXILINXFirmware(sc);
	
	if (ueagle_boot(sc))
		goto cleanup;

	/*
	 * Setup our /proc interface
	 */
	snprintf(path, 32, "%03d-%03d", usb->bus->busnum, usb->devnum);

	procfile =
	    create_proc_read_entry(path, 0, eu_procdir, eu_read_proc, sc);

	if (procfile) {
		procfile->owner = THIS_MODULE;
		eu_report("created proc entry at : /proc/driver/eagle-usb/%s\n",
			  path);
	} else {
		eu_err("failed to create proc entry at : /proc/driver/%s !\n",
		       path);
	}
	/* sysfs interface */
	device_create_file(&intf->dev, &dev_attr_status);
	device_create_file(&intf->dev, &dev_attr_mflags);

	eu_leaves(DBG_INIT);
	return 0;
      cleanup:
	kfree(sc);
	return -1;
}

static void eu_unbind(struct usbatm_data *usbatm, struct usb_interface *intf)
{
	char path[32];
	struct usb_device *usb = interface_to_usbdev(intf);
	struct ueagle_softc *sc = (struct ueagle_softc *)usbatm->driver_data;

	/*
	 * Free memory we alloced
	 */

	ueagle_stop(sc);

	snprintf(path, 32, "%03d-%03d", usb->bus->busnum, usb->devnum);
	remove_proc_entry(path, eu_procdir);
	/* sysfs interface */
	device_remove_file(&intf->dev, &dev_attr_status);
	device_remove_file(&intf->dev, &dev_attr_mflags);

	kfree(sc);
	eu_report("ADSL device removed\n");
}

static struct usbatm_driver euatm_drv = {
	.driver_name = "ueagle-atm",
	.owner = THIS_MODULE,
	.bind = eu_bind,
	.atm_start = eu_atm_open,
	.unbind = eu_unbind,
	.atm_stop = eu_atm_stop,
	.heavy_init = eu_heavy,
	.in = EP_BULK_DATA_OUT,.out = EP_BULK_DATA_OUT,
};
static int eu_probe(struct usb_interface *intf, const struct usb_device_id *id)
{
	struct usb_device *usb = interface_to_usbdev(intf);

	eu_enters(DBG_INIT);
	eu_dbg(DBG_INIT, "vid (%#X) pid (%#X) \n",
	       usb->descriptor.idVendor, usb->descriptor.idProduct);

	if (!(id->driver_info & PSTFIRM)) {
		return eu_load_firmware(usb, id->driver_info & 0xf);
	}

	if (dev >= NB_MODEM)
		return -ENODEV;

	usb_reset_device(usb);

	if (iso[dev]) {
#ifdef UDSL_DEFAULT_RCV_BUF_PER_URB
		//XXX try other alternate insterface it failed
		/*
		 * Set alternate interface to 8, which is ISO transport
		 * with the max. packet size ( about 1007 bytes)
		 */
		if (usb_set_interface(usb, USB_INTF_IN, FASTEST_ISO_INTF) < 0) {
			eu_err("usb_set_interface failed on iso alt 8\n");
		}
		else {
			euatm_drv.in = EP_ISOC_DATA_IN;
			euatm_drv.iso_frame_size = FASTEST_ISO_RATE;
			eu_report("iso mode\n");
		}
#else
#warning usbatm have no iso patch
		eu_report("iso mode not supported\n");
#endif
	}

	return usbatm_usb_probe(intf, id, &euatm_drv);
}

static void eu_disconnect(struct usb_interface *intf)
{
	struct usb_device *usb = interface_to_usbdev(intf);
	int ifnum = intf->altsetting->desc.bInterfaceNumber;
	eu_enters(DBG_INIT);
	if (usb->config->desc.bNumInterfaces == 3 && ifnum == 0)
		usbatm_usb_disconnect(intf);
	
	eu_leaves(DBG_INIT);
}

/**
 * eu_read_proc  - /proc interface.
 *
 */
static int eu_read_proc(char *page,
			char **start,
			off_t off, int count, int *eof, void *data)
{
	struct ueagle_softc *sc = (struct ueagle_softc *)data;
	int size;
	char *p = page;
	/*
	 * If this output needs to be larger than 4K (PAGE_SIZE), we need to do this
	 * differently
	 */
	p += sprintf(p, "eagle-usb status display\n");
	p += sprintf(p,
		     "-------------------------------------------------------------\n");
	p += sprintf(p, "Driver version %s     Chipset: Eagle%1.1d\n",
		     EAGLEUSBVERSION, sc->driver_info&0xf);

	if (sc->sc_dev != NULL) {
		p += sprintf(p,
			     "Vendor ID : 0x%x     Product ID : 0x%x   Rev: 0x%x\n",
			     sc->sc_dev->descriptor.idVendor,
			     sc->sc_dev->descriptor.idProduct,
			     sc->sc_dev->descriptor.bcdDevice);
		p += sprintf(p,
			     "USB Bus : %03d\t USB Device : %03d\t Dbg mask: 0x%x\n",
			     sc->sc_dev->bus->busnum, sc->sc_dev->devnum,
			     module_dbg_mask);
	}

	p += sprintf(p, "Tx Rate   %10d  Rx Rate   %10d\n",
		     sc->stats.phy.usrate, sc->stats.phy.dsrate);
	p += sprintf(p, "Tx Margin %10d  Rx Margin %10d   Atten   %10d dB\n",
		     sc->stats.phy.usmargin,
		     sc->stats.phy.dsmargin, sc->stats.phy.attenuation);
	p += sprintf(p, "Tx error  %10d  Rx error  %10d\n",
		     sc->stats.phy.userror, sc->stats.phy.dserror);
	p += sprintf(p, "Tx FEC    %10d  Rx FEC    %10d\n",
		     sc->stats.phy.usunc, sc->stats.phy.dsunc);
	p += sprintf(p, "Tx flow   %10d  Rx flow   %10d\n",
		     sc->stats.phy.txflow, sc->stats.phy.rxflow);
	p += sprintf(p, "VID-CPE   %10d  VID-CO    %10d    HEC    %10d\n",
		     sc->stats.phy.vidcpe,
		     sc->stats.phy.vidco, sc->stats.phy.cells_crc_errors);

	p += sprintf(p, "Delin          ");
	/*Delineation is the only one where we print a string sctead of a number */
	if (sc->stats.phy.flags & 0x0C00)
		p += sprintf(p, "ERROR");
	else if (sc->stats.phy.flags & 0x0030)
		p += sprintf(p, " LOSS");
	else
		p += sprintf(p, " GOOD");

	p += sprintf(p, "  Flags     %10.10x\n",sc->stats.phy.flags);
	switch ((sc->stats.phy.state >> 8) & 0xf) {
	case 1:
		p += sprintf(p, "Modem is initializing\n");
		break;
	case 0:
		p += sprintf(p, "Modem is booting\n");
		break;
	case 2:
		p += sprintf(p, "Modem is operational\n");
		break;
	default:
		p += sprintf(p, "Modem synchronization failed\n");
		break;
	}
	p += sprintf(p, "\n");
	/*
	 * Figure out how much of the page we used
	 */
	size = p - page;
	size -= off;

	if (size < count) {
		*eof = 1;
		if (size <= 0)
			return 0;
	} else
		size = count;
	/*
	 * Fool caller into thinking we started where he told us to in the page
	 */
	*start = page + off;
	return size;
}



/*
 * List of supported VID/PID
 */
static const struct usb_device_id eu_ids[] = {
	{USB_DEVICE(ELSA_VID, ELSA_PID_PREFIRM),.driver_info =
	 ELSA | PREFIRM},
	{USB_DEVICE(ELSA_VID, ELSA_PID_PSTFIRM),.driver_info =
	 ELSA | PSTFIRM},
	{USB_DEVICE(EAGLE_VID, EAGLE_I_PID_PREFIRM),.driver_info =
	 EAGLE_I | PREFIRM},
	{USB_DEVICE(EAGLE_VID, EAGLE_I_PID_PSTFIRM),.driver_info =
	 EAGLE_I | PSTFIRM},
	{USB_DEVICE(EAGLE_VID, EAGLE_II_PID_PREFIRM),.driver_info =
	 EAGLE_II | PREFIRM},
	{USB_DEVICE(EAGLE_VID, EAGLE_II_PID_PSTFIRM),.driver_info =
	 EAGLE_II | PSTFIRM},
	{USB_DEVICE(EAGLE_VID, EAGLE_IIC_PID_PREFIRM),.driver_info =
	 EAGLE_II | PREFIRM},
	{USB_DEVICE(EAGLE_VID, EAGLE_IIC_PID_PSTFIRM),.driver_info =
	 EAGLE_II | PSTFIRM},
	{USB_DEVICE(EAGLE_VID, EAGLE_III_PID_PREFIRM),.driver_info =
	 EAGLE_III | PREFIRM},
	{USB_DEVICE(EAGLE_VID, EAGLE_III_PID_PSTFIRM),.driver_info =
	 EAGLE_III | PSTFIRM},
	{USB_DEVICE(USR_VID, MILLER_A_PID_PREFIRM),.driver_info =
	 EAGLE_I | PREFIRM},
	{USB_DEVICE(USR_VID, MILLER_A_PID_PSTFIRM),.driver_info =
	 EAGLE_I | PSTFIRM},
	{USB_DEVICE(USR_VID, MILLER_B_PID_PREFIRM),.driver_info =
	 EAGLE_I | PREFIRM},
	{USB_DEVICE(USR_VID, MILLER_B_PID_PSTFIRM),.driver_info =
	 EAGLE_I | PSTFIRM},
	{USB_DEVICE(USR_VID, HEINEKEN_A_PID_PREFIRM),.driver_info =
	 EAGLE_I | PREFIRM},
	{USB_DEVICE(USR_VID, HEINEKEN_A_PID_PSTFIRM),.driver_info =
	 EAGLE_I | PSTFIRM},
	{USB_DEVICE(USR_VID, HEINEKEN_B_PID_PREFIRM),.driver_info =
	 EAGLE_I | PREFIRM},
	{USB_DEVICE(USR_VID, HEINEKEN_B_PID_PSTFIRM),.driver_info =
	 EAGLE_I | PSTFIRM},
	{}
};

/*
 * USB driver descriptor
 */
static struct usb_driver eu_driver = {
	.owner = THIS_MODULE,
	.name = "ueagle-atm",
	.id_table = eu_ids,
	.probe = eu_probe,
	.disconnect = eu_disconnect,
};

MODULE_DEVICE_TABLE(usb, eu_ids);

/**
 * eu_init - Initialize the module.
 *      Creates /proc/driver/eagle-usb directory
 *      Register to USB subsystem
 *
 */
static int __init eu_init(void)
{
	int result = 0;

	eu_enters(DBG_INIT);

	eu_report("driver V" EAGLEUSBVERSION " loaded\n");

	eu_procdir = proc_mkdir("driver/ueagle-atm", NULL);

	if (eu_procdir) {
		eu_procdir->owner = THIS_MODULE;
	} else {
		eu_report("could not create /proc/driver/eagle-usb/\n");
		result = -ENOMEM;
	}

	usb_register(&eu_driver);

	eu_leaves(DBG_INIT);

	return 0;
}

module_init(eu_init);

/**
 * eu_exit  -  Finalize module
 *
 *    Deregister with USB subsystem
 *    Remove /proc/drivers/eagle-usb directory
 *
 */
static void __exit eu_exit(void)
{
	eu_enters(DBG_INIT);

	/*
	 * This calls automatically the eu_disconnect method if necessary:
	 */
	usb_deregister(&eu_driver);

	eu_report("driver unloaded\n");
	remove_proc_entry("driver/ueagle-atm", NULL);

	eu_leaves(DBG_INIT);

}

module_exit(eu_exit);

MODULE_AUTHOR("Frederick Ros/Matthieu Castet/Damien Bergamini");
MODULE_DESCRIPTION("Eagle USB ADSL Modem driver");
MODULE_LICENSE("GPL");
